<?php

/*
 * Core dmemcache functions required by:
 *   memcache_memcache.inc
 *   memcache_memcache.db.inc
 *   memcache_session-memcache.inc
 *   memcache_session-memcache.db.inc
 */

global $_demandbase_memcache_statistics;
$_demandbase_memcache_statistics = array();

/*
 * A memcache API for Drupal.
 */

/**
 *  Place an item into memcache
 *
 * @param $key The string with with you will retrieve this item later.
 * @param $value The item to be stored.
 * @param $exp Parameter expire is expiration time in seconds. If it's 0, the item never expires
 *   (but memcached server doesn't guarantee this item to be stored all the time, it could be
 *   deleted from the cache to make place for other items).
 * @param $bin The name of the Drupal subsystem that is making this call. Examples could be
 *   'cache', 'alias', 'taxonomy term' etc. It is possible to map different $bin values to
 *   different memcache servers.
 *
 * @return bool
 */
function demandbase_dmemcache_set($key, $value, $exp = 0, $bin = 'cache') {
  global $_demandbase_memcache_statistics;
  $full_key = demandbase_dmemcache_key($key, $bin);
  $success = FALSE;
  if ($mc = demandbase_dmemcache_object($bin)) {
    if ($mc instanceof Memcached) {
      $success = $mc->set($full_key, $value, $exp);
    }
    elseif ($mc instanceof Memcache) {
      $success = $mc->set($full_key, $value, MEMCACHE_COMPRESSED, $exp);
    }
  }
  $_demandbase_memcache_statistics[] = array('set', $bin, $full_key, (int)$success);
  return $success;
}

/**
 *  Add an item into memcache
 *
 * @param $key The string with which you will retrieve this item later.
 * @param $value The item to be stored.
 * @param $exp Parameter expire is expiration time in seconds. If it's 0, the
 *   item never expires (but memcached server doesn't guarantee this item to be
 *   stored all the time, it could be deleted from the cache to make place for
 *   other items).
 * @param $bin The name of the Drupal subsystem that is making this call.
 *   Examples could be 'cache', 'alias', 'taxonomy term' etc. It is possible
 *   to map different $bin values to different memcache servers.
 * @param $mc Optionally pass in the memcache object.  Normally this value is
 *   determined automatically based on the bin the object is being stored to.
 * @param $flag If using the older memcache PECL extension as opposed to the
 *   newer memcached PECL extension, the MEMCACHE_COMPRESSED flag can be set
 *   to use zlib to store a compressed copy of the item.  This flag option is
 *   completely ignored when using the newer memcached PECL extension.
 *
 * NOTE: demandbase_dmemcache_get calls demandbase_dmemcache_add when grabbing a semaphore.  For that
 * reason, demandbase_dmemcache_add can't call demandbase_dmemcache_get for any reason or you'll end
 * up in an infinite loop.
 *
 * @return bool
 */
function demandbase_dmemcache_add($key, $value, $exp = 0, $bin = 'cache', $mc = NULL, $flag = FALSE) {
  global $_demandbase_memcache_statistics;
  $full_key = demandbase_dmemcache_key($key, $bin);
  $success = FALSE;
  if ($mc || ($mc = demandbase_dmemcache_object($bin))) {
    if ($mc instanceof Memcached) {
      $success = $mc->add($full_key, $value, $exp);
    }
    elseif ($mc instanceof Memcache) {
      $success = $mc->add($full_key, $value, $flag, $exp);
    }
  }
  $_demandbase_memcache_statistics[] = array('add', $bin, $full_key, (int)$success);
  return $success;
}

/**
 * Retrieve a value from the cache.
 *
 * @param $key The key with which the item was stored.
 * @param $bin The bin in which the item was stored.
 *
 * @return The item which was originally saved or FALSE
 */
function demandbase_dmemcache_get($key, $bin = 'cache') {
  global $_demandbase_memcache_statistics;
  $full_key = demandbase_dmemcache_key($key, $bin);
  $statistics = array('get', $bin, $full_key);
  $success = '0';
  if ($mc = demandbase_dmemcache_object($bin)) {
    $track_errors = ini_set('track_errors', '1');
    $php_errormsg = '';

    $result = @$mc->get($full_key);
    $statistics[] = (bool) $result;
    $_demandbase_memcache_statistics[] = $statistics;

    if (!empty($php_errormsg)) {
      register_shutdown_function('watchdog', 'memcache', 'Exception caught in demandbase_dmemcache_get: !msg', array('!msg' => $php_errormsg), WATCHDOG_WARNING);
      $php_errormsg = '';
    }
    ini_set('track_errors', $track_errors);

    return $result;
  }
}

/**
 * Retrieve multiple values from the cache.
 *
 * @param $keys The keys with which the items were stored.
 * @param $bin The bin in which the item was stored.
 *
 * @return The item which was originally saved or FALSE
 */
function demandbase_dmemcache_get_multi($keys, $bin = 'cache', $mc = NULL) {
  global $_demandbase_memcache_statistics;
  $full_keys = array();
  $statistics = array();
  foreach ($keys as $key => $cid) {
    $full_key = demandbase_dmemcache_key($cid, $bin);
    $statistics[$full_key] = array('getMulti', $bin, $full_key);
    $full_keys[] = $full_key;
  }
  $results = array();
  if ($mc || ($mc = demandbase_dmemcache_object($bin))) {
    if ($mc instanceof Memcached) {
      $results = $mc->getMulti($full_keys);
    }
    elseif ($mc instanceof Memcache) {
      $track_errors = ini_set('track_errors', '1');
      $php_errormsg = '';

      $results = @$mc->get($full_keys);

      if (!empty($php_errormsg)) {
        register_shutdown_function('watchdog', 'memcache', 'Exception caught in demandbase_dmemcache_get_multi: !msg', array('!msg' => $php_errormsg), WATCHDOG_WARNING);
        $php_errormsg = '';
      }
      ini_set('track_errors', $track_errors);
    }
  }
  foreach ($statistics as $key => $values) {
    $values[] = isset($results[$key]) ? '1': '0';
    $_demandbase_memcache_statistics[] = $values;
  }
  return $results;
}

/**
 * Deletes an item from the cache.
 *
 * @param $key The key with which the item was stored.
 * @param $bin The bin in which the item was stored.
 *
 * @return Returns TRUE on success or FALSE on failure.
 */
function demandbase_dmemcache_delete($key, $bin = 'cache') {
  global $_demandbase_memcache_statistics;
  $full_key = demandbase_dmemcache_key($key, $bin);
  $success = FALSE;
  if ($mc = demandbase_dmemcache_object($bin)) {
    $success = $mc->delete($full_key, 0);
  }
  $_demandbase_memcache_statistics[] = array('delete', $bin, $full_key, (int)$success);
  return $success;
}

/**
 * Immediately invalidates all existing items. demandbase_dmemcache_flush doesn't actually free any
 * resources, it only marks all the items as expired, so occupied memory will be overwritten by
 * new items.
 *
 * @param $bin The bin to flush. Note that this will flush all bins mapped to the same server
 *   as $bin. There is no way at this time to empty just one bin.
 *
 * @return Returns TRUE on success or FALSE on failure.
 */
function demandbase_dmemcache_flush($bin = 'cache') {
  global $_demandbase_memcache_statistics;
  $_demandbase_memcache_statistics[] = array('flush', $bin, '', '');
  if ($mc = demandbase_dmemcache_object($bin)) {
    return $mc->flush();
  }
}

function demandbase_dmemcache_stats($stats_bin = 'cache', $stats_type = 'default', $aggregate = FALSE) {
  $memcache_bins = variable_get('memcache_bins', array('cache' => 'default'));
  // The stats_type can be over-loaded with an integer slab id, if doing a
  // cachedump.  We know we're doing a cachedump if $slab is non-zero.
  $slab = (int)$stats_type;

  foreach ($memcache_bins as $bin => $target) {
    if ($stats_bin == $bin) {
      if ($mc = demandbase_dmemcache_object($bin)) {
        if ($mc instanceof Memcached) {
          $stats[$bin] = $mc->getStats();
        }
        // The PHP Memcache extension 3.x version throws an error if the stats
        // type is NULL or not in {reset, malloc, slabs, cachedump, items,
        // sizes}. If $stats_type is 'default', then no parameter should be
        // passed to the Memcache memcache_get_extended_stats() function.
        else if ($mc instanceof Memcache) {
          if ($stats_type == 'default' || $stats_type == '') {
            $stats[$bin] = $mc->getExtendedStats();
          }
          // If $slab isn't zero, then we are dumping the contents of a
          // specific cache slab.
          else if (!empty($slab))  {
            $stats[$bin] = $mc->getStats('cachedump', $slab);
          }
          else {
            $stats[$bin] = $mc->getExtendedStats($stats_type);
          }
        }
      }
    }
  }
  // Optionally calculate a sum-total for all servers in the current bin.
  if ($aggregate) {
    // Some variables don't logically aggregate.
    $no_aggregate = array('pid', 'time', 'version', 'pointer_size', 'accepting_conns', 'listen_disabled_num');
    foreach($stats as $bin => $servers) {
      if (is_array($servers)) {
        foreach ($servers as $server) {
          if (is_array($server)) {
            foreach ($server as $key => $value) {
              if (!in_array($key, $no_aggregate)) {
                if (isset($stats[$bin]['total'][$key])) {
                  $stats[$bin]['total'][$key] += $value;
                }
                else {
                  $stats[$bin]['total'][$key] = $value;
                }
              }
            }
          }
        }
      }
    }
  }
  return $stats;
}

/**
 * Returns an Memcache object based on the bin requested. Note that there is
 * nothing preventing developers from calling this function directly to get the
 * Memcache object. Do this if you need functionality not provided by this API
 * or if you need to use legacy code. Otherwise, use the dmemcache (get, set,
 * delete, flush) API functions provided here.
 *
 * @param $bin The bin which is to be used.
 *
 * @param $flush Rebuild the bin/server/cache mapping.
 *
 * @return an Memcache object or FALSE.
 */
function demandbase_dmemcache_object($bin = NULL, $flush = FALSE) {
  static $extension, $memcacheCache = array(), $memcache_servers, $memcache_bins, $memcache_persistent;

  if (!isset($extension)) {
    // If an extension is specified in settings.php, use that when available.
    $preferred = ucfirst(strtolower((variable_get('memcache_extension', NULL))));
    if (isset($preferred) && class_exists($preferred)) {
      $extension = $preferred;
    }
    // If no extension is set, default to Memcache.
    // The Memcached extension has some features that the older extension lacks
    // but also an unfixed bug that affects cache clears.
    // @see http://pecl.php.net/bugs/bug.php?id=16829
    elseif (class_exists('Memcache')) {
      $extension = 'Memcache';
    }
    elseif (class_exists('Memcached')) {
      $extension = 'Memcached';
    }
  }

  if ($flush) {
    foreach ($memcacheCache as $cluster) {
      $cluster->close();
    }
    $memcacheCache = array();
  }

  if (empty($memcacheCache) || empty($memcacheCache[$bin])) {
    // $memcache_servers and $memcache_bins originate from settings.php.
    // $memcache_servers_custom and $memcache_bins_custom get set by
    // memcache.module. They are then merged into $memcache_servers and
    // $memcache_bins, which are statically cached for performance.
    if (empty($memcache_servers)) {
      // Values from settings.php
      $memcache_servers = variable_get('memcache_servers', array('127.0.0.1:11211' => 'default'));
      $memcache_bins    = variable_get('memcache_bins', array('cache' => 'default'));
    }

    // If there is no cluster for this bin in $memcache_bins, cluster is 'default'.
    $cluster = empty($memcache_bins[$bin]) ? 'default' : $memcache_bins[$bin];

    // If this bin isn't in our $memcache_bins configuration array, and the
    // 'default' cluster is already initialized, map the bin to 'cache' because
    // we always map the 'cache' bin to the 'default' cluster.
    if (empty($memcache_bins[$bin]) && !empty($memcacheCache['cache'])) {
      $memcacheCache[$bin] = &$memcacheCache['cache'];
    }
    else {
      // Create a new Memcache object. Each cluster gets its own Memcache
      // object.
      if ($extension == 'Memcached') {
        $memcache = new Memcached;
        $default_opts = array(
          Memcached::OPT_COMPRESSION => FALSE,
          Memcached::OPT_DISTRIBUTION => Memcached::DISTRIBUTION_CONSISTENT,
        );
        foreach ($default_opts as $key => $value) {
          $memcache->setOption($key, $value);
        }
        // See README.txt for setting custom Memcache options when using the
        // memcached PECL extension.
        $memconf = variable_get('memcache_options', array());
        foreach ($memconf as $key => $value) {
          $memcache->setOption($key, $value);
        }
      }
      elseif ($extension == 'Memcache') {
        $memcache = new Memcache;
      }
      else {
        drupal_set_message('You must enable the PECL memcached or memcache extension to use memcache.inc.', 'error');
        return;
      }

      // Indicate whether to connect to memcache using a persistent connection.
      // Note: this only affects the Memcache PECL extension, and does not
      // affect the Memcached PECL extension.  For a detailed explanation see:
      //  http://drupal.org/node/822316#comment-4427676
      if (!isset($memcache_persistent)) {
        $memcache_persistent = variable_get('memcache_persistent', FALSE);
      }

      // A variable to track whether we've connected to the first server.
      $init = FALSE;

      // Link all the servers to this cluster.
      foreach ($memcache_servers as $s => $c) {
        if ($c == $cluster) {
          list($host, $port) = explode(':', $s);

          // Support unix sockets in the format 'unix:///path/to/socket'.
          if ($host == 'unix') {
            // When using unix sockets use the full path for $host.
            $host = $s;
            // Port is always 0 for unix sockets.
            $port = 0;
          }

          // Using the Memcache PECL extension.
          if ($memcache instanceof Memcache) {
            // When using the PECL memcache extension, we must use ->(p)connect
            // for the first connection.
            if (!$init) {
              $track_errors = ini_set('track_errors', '1');
              $php_errormsg = '';

              if ($memcache_persistent && @$memcache->pconnect($host, $port)) {
                $init = TRUE;
              }
              elseif (!$memcache_persistent && @$memcache->connect($host, $port)) {
                $init = TRUE;
              }

              if (!empty($php_errormsg)) {
                register_shutdown_function('watchdog', 'memcache', 'Exception caught in demandbase_dmemcache_object: !msg', array('!msg' => $php_errormsg), WATCHDOG_WARNING);
                $php_errormsg = '';
              }
              ini_set('track_errors', $track_errors);
            }
            else {
              $memcache->addServer($host, $port, $memcache_persistent);
            }
          }
          // Using the Memcached PECL extension.
          else if ($memcache->addServer($host, $port) && !$init) {
            $init = TRUE;
          }
        }
      }

      if ($init) {
        // Map the current bin with the new Memcache object.
        $memcacheCache[$bin] = $memcache;

        // Now that all the servers have been mapped to this cluster, look for
        // other bins that belong to the cluster and map them too.
        foreach ($memcache_bins as $b => $c) {
          if ($c == $cluster && $b != $bin) {
            // Map this bin and cluster by reference.
            $memcacheCache[$b] = &$memcacheCache[$bin];
          }
        }
      }
    }
  }

  return empty($memcacheCache[$bin]) ? FALSE : $memcacheCache[$bin];
}

function demandbase_dmemcache_key($key = '', $bin = 'cache', $reset = FALSE) {
  static $prefix;
  if (!isset($prefix)) {
    $prefix = '';
    // memcache_key_prefix can be set in settings.php to support site namespaces
    // in a multisite environment.
    if ($prefix = variable_get('memcache_key_prefix', '')) {
      $prefix .= '-';
    }
    // When simpletest is running, emulate the simpletest database prefix here
    // to avoid the child site setting cache entries in the parent site.
    if (isset($GLOBALS['drupal_test_info']['test_run_id'])) {
      $prefix .= $GLOBALS['drupal_test_info']['test_run_id'];
    }
  }
  $key = demandbase_dmemcache_change_key($key); // Change key for anonymous users
  $full_key = urlencode($prefix . $bin . '-' . $key);

  // Memcache only supports key lengths up to 250 bytes.  If we have generated
  // a longer key, hash it with sha1 which will shrink the key down to 40 bytes
  // while still keeping it unique.
  if (strlen($full_key) > 250) {
    $full_key = $prefix . $bin . '-' . sha1($key);
  }

  return $full_key;
 }

/**
 *
 * Change key for anonymous users
 * Also check whether this feature need to be activated.
 * 
 * @param String $key
 */
function demandbase_dmemcache_change_key($key) {
  global $user;
  if (!(isset($user->uid) && ($user->uid == 0))) {
    return $key;
  }
  $activate_for_anonymous = variable_get('demandbase_activate_cache_for_anonymous', '0');
  if ($activate_for_anonymous != 1) {
    return $key;
  }
  $activate_memcache_paths = variable_get('demandbase_activate_cache_path_list', '0');
  if (!$activate_memcache_paths) {
    return $key . '_' . ip_address();
  }
  $path_list = variable_get('demandbase_cache_path_list', '');
  $paths = explode("\n", $path_list);     
  if (empty($path_list) ){
    return $key;
  }
  $path_matched = demandbase_dmemcache_validate_path($paths);
  if ($path_matched) {
     $key = $key . '+' . ip_address();
  }
  return $key;
}

/**
 * 
 * Enter description here ...
 * @param unknown_type $pages
 */
function demandbase_dmemcache_validate_path($pages) {
  $path = demandbase_dmemcache_get_path_alias($_GET['q']);
   // Compare with the internal and path alias (if any).
   $page_match = demandbase_dmemcache_match($path, $pages, TRUE);
   if ($path != $_GET['q']) {
     $page_match = $page_match || demandbase_dmemcache_match($_GET['q'], $pages, TRUE);
   }
  return $page_match;
}

/**
 * 
 * Enter description here ...
 * @param unknown_type $path
 */
function demandbase_dmemcache_get_path_alias($path) {
  // @TODO implement path alias feature
  return $path;
}

/**
 * 
 * Enter description here ...
 * @param String $subject
 * @param Array $patterns
 * @param Boolean $path
 */
function demandbase_dmemcache_match($subject, $patterns, $path = FALSE) {
    static $regexps;
    $match = FALSE;
    $positives = $negatives = 0;
    $subject = !is_array($subject) ? array($subject) : $subject;
    foreach ($patterns as $pattern) {
      if (strpos($pattern, '~') === 0) {
        $negate = TRUE;
        $negatives++;
      }
      else {
        $negate = FALSE;
        $positives++;
      }
      $pattern = ltrim($pattern, '~');
      if (!isset($regexps[$pattern])) {
        if ($path) {
          $regexps[$pattern] = '/^(' . preg_replace(array('/(\r\n?|\n)/', '/\\\\\*/', '/(^|\|)\\\\<front\\\\>($|\|)/'), array('|', '.*', '\1' . preg_quote(variable_get('site_frontpage', 'node'), '/') . '\2'), preg_quote($pattern, '/')) . ')$/';
        }
        else {
          $regexps[$pattern] = '/^(' . preg_replace(array('/(\r\n?|\n)/', '/\\\\\*/'), array('|', '.*'), preg_quote($pattern, '/')) . ')$/';
        }
      }
      foreach ($subject as $value) {
        if (preg_match($regexps[$pattern], $value)) {
          if ($negate) {
            return FALSE;
          }
          $match = TRUE;
        }
      }
    }
    // If there are **only** negative conditions and we've gotten this far none
    // we actually have a match.
    if ($positives === 0 && $negatives) {
      return TRUE;
    }
    return $match;
  }
